unit dsObjAngle;
{

TAngle : verision 0.91 8/7/99
WHY?
hmmmm, well, I'm a land surveyor (geomatist / geomatician if you like) and got
sick and tired of continually messing around with angular input in my
Delphi programs.

DISCLAIMER
First of all, don't laugh, this is my first attempt.

Free to one and all, do whatever you want, with whatever you want with this unit,
but accept it as is. I take no responsibility for any damage to anything ever
caused by using this unit, especially nuclear reactors and commercial aircraft!!!!!
If you find any bugs or can suggest any improvements to this 'ere thing, give us
a shout on -

     danky@bigfoot.com

     or check out latest releases on

     http://www.pookaloo.f9.co.uk/prog

0.951
(i) Three constructors implemented, one each for DMS, Decimal or Radian input, rather
than just having the constructor allocating and then assigning data to the angle.

(ii) Whoops, constructor was declared as a protected rather than a public method ...
now corrected.

0.952
(i) Additional ancilliary functions added for trignometrical calculations:

    a) Bearing Calculation
    b) Horizontal Distance Calculation
    c) Delta Easting and Delta Northing Calulations
    d) Easting and Northing Calulations

0.953
(i) Altered SinAngle, CosAngle and TanAngle methods to compute radian stylee data
rather than converting deg, min & sec again.

(ii) Added 3D Coordinate Object (TCoordinate) for further survey work computations.
Nothing special, just checks for appropriate data on object creation.

0.954
(i) Added new constructor to create new angle (really a bearing or azimuth) based
on difference in Easting (or X) and difference in Northing (Y) string input.

(ii) Whoops, sorry, added ds to the front end of the unit name to keep consistncy with
other units I'm developing
copyright (c) 1999 Dankysoft

}

interface

uses
 Classes, SysUtils, Math;

const
 pi180 = pi/180;

type

 EAngleOutOfRange = class(Exception);

 TSign = (Positive, Negative);

 TAngle = class(TObject)
 private
  FDegrees, FMinutes : integer;
  FSeconds : single;
 protected
  procedure SetDMSText(value: string);
  function GetDMSText: string;
  procedure SetDecimalText(value: string);
  function GetDecimalText: string;
  procedure SetRadianText(value: string);
  function GetRadianText: string;
  procedure SetDegrees(value: integer);
  procedure SetMinutes(value: integer);
  procedure SetSeconds(value: single);
 public
  constructor CreateDMS(value: string);
  constructor CreateDecimal(value: string);
  constructor CreateRadian(value: string);
  constructor CreateDiffCoords(DeltaEasting, DeltaNorthing : string);
  function SinAngle: double;
  function CosAngle: double;
  function TanAngle: double;
  property Degrees : integer read FDegrees write SetDegrees;
  property Minutes : integer read FMinutes write SetMinutes;
  property Seconds : single read FSeconds write SetSeconds;
  property DMSText : string read GetDMSText write SetDMSText;
  property DecimalText : string read GetDecimalText write SetDecimalText;
  property RadianText : string read GetRadianText write SetRadianText;
 end;

 TCoordinate = class(TObject)
 private
  FEasting, FNorthing, FHeight : Double;
 protected
 published
  constructor CreateXYZ(eValue, nValue, hValue : string);
  property Easting : Double read FEasting write FEasting;
  property Northing : Double read FNorthing write FNorthing;
  property Height : Double read FHeight write FHeight;
 end;

function DecodeString(SearchStr: string; var BigString: string):string;
function CheckAngleRange(DecimalAngle : string): string;
function CheckInputFormat(DMSAngle : string):string;
function BearingCalculation(dE, dN : Double): string;
function HzDistCalc(dEasting, dNorthing : double): string;
function deltaEastingCalc(DecimalBearing, Distance : double): double;
function deltaNorthingCalc(DecimalBearing, Distance : Double): double;
function EastingCalc(DecimalBearing, Distance, AtEasting : double): double;
function NorthingCalc(DecimalBearing, Distance, AtNorthing : double): double;

implementation

{********* Ancilliary trignometrical functions ********}

function DecodeString(SearchStr: string; var BigString: string):string;
var p : integer;
    blank : boolean;
begin
try
  blank := true;
    while blank do
    begin
       blank := (pos(SearchStr, BigString) = 1);
       if blank then delete(BigString, 1, 1);
    end;
  p:= pos(SearchStr, BigString);
  if p = 0 then result := bigstring
  else if ((p <25000) and (p>0)) then
  begin
    result := copy(BigString, 1, p-1);
    delete(BigString, 1, p);
  end;
 except
  raise EConvertError.create('Unable to convert data');
  exit;
 end;
end;

// Checks, reduces & returns decimal angle values, so that they are
//within 0 to 360 degs. Same as SurvCalc's BearingSizeCheck
function CheckAngleRange(DecimalAngle: string): string;
var dec : double;
begin
 try
  dec := StrToFloat(DecimalAngle);
 except
  raise EConvertError.create(DecimalAngle + ' is not a valid decimal angle');
  exit;
 end;
 if (dec > 10800) or (dec < -10800) then
 begin
  raise
  EConvertError.create('Please enter a sensible sized angle');
  exit;
 end;
 if dec >= (360) then while (dec >= (360)) do dec := dec - (360)
 else if dec < 0 then while (dec < 0) do dec := dec + (360)
 else dec := dec;
 Result := FloatToStr(dec);
end;

// Checks to see whether angle is in form of ddd.mmss or ddd.mm.ss,
// all methods assume ddd.mmss format
function CheckInputFormat(DMSAngle: string): string;
var c, i : integer;
    str, d, m, s : string;
begin
 c := 0;
 i := 1;
 str := DMSAngle;
 while i <> 0 do
 begin
  i := pos('.', str);
  delete(str, 1, i);
  if i <> 0 then c := c + 1;
 end;

 if c > 1 then
 begin
  d := DecodeString('.', DMSAngle);
  m := DecodeString('.', DMSAngle);
  s := Trim(DMSAngle);
  result := d + '.' + m + s;
 end
 else result := DMSAngle;
end;

//Calulates Bearing from easting difference and northing differences
//returns a decimal degree string
function BearingCalculation(dE, dN : Double): string;
var Brg : double;
    Ang : TAngle;
begin
  if dN = 0 then
     if dE >= 0 then Brg := (pi / 2) else Brg := (3 * (pi / 2))
  else if dE = 0 then
     if dN >= 0 then Brg := 0 else Brg := pi
  else
  begin
    Brg := ArcTan(dE / dN);
    if dE > 0 then
      if dN > 0 then Brg := Brg else Brg := Brg + pi
    else if dE < 0 then
      if dN > 0 then Brg := BRg + (2 * pi) else Brg := Brg + pi
  end;
  Ang := TAngle.CreateRadian(FloatToStr(Brg));
  Result := CheckAngleRange(Ang.DecimalText);
  Ang.Free;
end;

//Calculates horizontal distance from two points coordinates
function HzDistCalc(dEasting, dNorthing : double): string;
begin
   result:= FormatFloat('0.000', sqrt(sqr(dEasting) + sqr(dNorthing)));
end;

//Calculates difference in Easting from a bearing & distance
function deltaEastingCalc(DecimalBearing, Distance : double): double;
var  bg : TAngle;
begin
  bg := TAngle.CreateDecimal(FloatToStr(DecimalBearing));
  result := Distance * Sin(StrToFloat(bg.RadianText));
  bg.Free;
end;

//Calculates difference in Northing from a bearing & distance
function deltaNorthingCalc(DecimalBearing, Distance : Double): double;
var  bg : TAngle;
begin
  bg := TAngle.CreateDecimal(FloatToStr(DecimalBearing));
  result := Distance * Cos(StrToFloat(bg.RadianText));
  bg.Free;
end;

//Calculates new Easting from a old Easting, bearing & distance
//takes bearing input as a radian value
function EastingCalc(DecimalBearing, Distance, AtEasting : double): double;
begin
  result := AtEasting + deltaEastingCalc(DecimalBearing, Distance);
end;

//Calculates new Northing from a old Northing, bearing & distance
//takes bearing input as a radian value
function NorthingCalc(DecimalBearing, Distance, AtNorthing : double): double;
begin
  result := AtNorthing + deltaNorthingCalc(DecimalBearing, Distance);
end;

{*********Start of TAngle methods**********************}

// Creates the Angle object with degrees, minutes & seconds input
constructor TAngle.CreateDMS(value: string);
begin
  SetDMSText(value);
end;

// Creates the Angle object with decimal degrees input
constructor TAngle.CreateDecimal(value: string);
begin
  SetDecimalText(value);
end;

// Creates the Angle object with radians input
constructor TAngle.CreateRadian(value: string);
begin
  SetRadianText(value);
end;

//Creates the Angle (Bearing?) object from difference in coordinate input
constructor TAngle.CreateDiffCoords(DeltaEasting, DeltaNorthing : string);
begin
  SetDecimalText(BearingCalculation(StrToFloat(DeltaEasting), StrToFloat(DeltaNorthing)));
end;

//Reads text input in and establishs values for ddd, mm and ss
procedure TAngle.SetDMSText(value : string);
var DMS : double;
    i, deg, min: integer;
    sec: single;
    sign : TSign;
begin
 value := CheckInputFormat(value);
 try
  DMS := StrToFloat(value);
 except
  raise EConvertError.create(value +' is not a valid deg/min/sec angle');
  exit;
 end;
  if DMS < 0 then sign := negative else sign := positive;
  DMS := Abs(DMS);
 try
  value := FormatFloat('##0.00000',DMS);
  i := pos('.', value);

  deg := StrToInt(copy(value, 1, i-1));
  min := StrToInt(copy(value, i+1, 2));
  sec := StrToFloat(copy(value, i+3, 3))/10;

  while deg >= 360 do deg := deg - 360;
  case sign of
   positive :
    begin
     SetDegrees(deg);
     SetMinutes(min);
     SetSeconds(sec);
    end;
   negative :
    begin
     if sec = 0 then sec := 0.001;
     SetDegrees(359-deg);
     SetMinutes(59-min);
     SetSeconds(60-sec);
    end;
  end;//case
 except
  on EAngleOutOfRange do
  begin
   FDegrees := 0;
   FMinutes := 0;
   FSeconds := 0;
   raise;
  end;
 end;
end;

// writes text ddd.mmss representation of angle values
function TAngle.GetDMSText: string;
begin
 result := FormatFloat('##0', Degrees) + '.' + FormatFloat('00', Minutes) + FormatFloat('00', Seconds);
end;

// reads degrees value, at this point, the value should already be 0 to 360 degs.
procedure TAngle.SetDegrees(value: integer);
begin
  FDegrees := value;
end;

// reads minutes value in
procedure TAngle.SetMinutes(value: integer);
begin
  if (value >= 0) and (value < 60) then FMinutes := value
  else raise EAngleOutOfRange.Create ('Minutes out of range');
end;

// reads seconds value in
procedure TAngle.SetSeconds(value: single);
begin
  if (value >= 0) and (value < 60) then FSeconds := value
  else raise EAngleOutOfRange.Create ('Seconds out of range');
end;

// calculates sine of objects angular value
function TAngle.SinAngle: double;
begin
  Result := Sin(StrToFloat(RadianText));
end;

// calculates cosine of objects angular value
function TAngle.CosAngle: double;
begin
// result := Cos(pi180 * (Degrees + (Minutes / 60) + (Seconds / 3600)));
  Result := Cos(StrToFloat(RadianText));
end;

// calculates tangent of objects angular value
function TAngle.TanAngle: double;
begin
  Result := Tan(StrToFloat(RadianText));
end;

// reads in angle values from decimal text input (may be a bit dodgey!!)
procedure TAngle.SetDecimalText(value: string);
var dec : double;
    sign : TSign;
    deg, min : integer;
    sec : single;
begin
 try
  dec := StrToFloat(value);
 except
  raise EConvertError.create(value +' is not a valid decimal angle');
  exit;
 end;
 if dec < 0 then sign := negative else sign := positive;
 dec := Abs(dec);
 //sort this bugger out
 try
  deg := Round(Int(dec));
  min := Round(Int(Int((dec-deg) * 60)));
  sec := (((dec-deg) * 60) - ((Int((dec-deg) * 60)))) * 60;
  while deg >= 360 do deg := deg - 360;
  case sign of
   positive :
    begin
     SetDegrees(deg);
     SetMinutes(min);
     SetSeconds(sec);
    end;
   negative :
    begin
     SetDegrees(359-deg);
     SetMinutes(59-min);
     SetSeconds(60-sec);
    end;
  end;//case
 except
  on EAngleOutOfRange do
  begin
   FDegrees := 0;
   FMinutes := 0;
   FSeconds := 0;
   raise;
  end;
 end;
end;

// writes current angular values out as decimal text value
function TAngle.GetDecimalText: string;
begin
 if Degrees < 0 then result := FloatToStr(Degrees - (Minutes / 60) - (Seconds / 3600))
  else result := FloatToStr(Degrees + (Minutes / 60) + (Seconds / 3600));
end;

// reads in angle values from radian text input (may be a bit dodgey!!)
procedure TAngle.SetRadianText(value: string);
var rad : double;
begin
 try
  rad := StrToFloat(value);
 except
  raise EConvertError.create(value +' is not a valid radian angle');
  exit;
 end;

  if (rad >= (2*pi)) or (rad <= (-2*pi)) then
    raise EAngleOutOfRange.Create ('Radians out of range')
  else SetDecimalText(FloatToStr (rad * (180 / pi)));

end;

// writes current angular values out as radian text value
function TAngle.GetRadianText: string;
begin
 result := FloatToStr(StrToFloat(GetDecimalText) * (pi / 180));
end;

//creates and allocates values to the coordinate object.
constructor TCoordinate.CreateXYZ(eValue, nValue, hValue : string);
begin
 try
  Easting := StrToFloat(eValue);
  Northing := StrToFloat(nValue);
  Height := StrToFloat(hValue);
 except
  raise EConvertError.Create('Input value is not valid.');
  exit;
 end;
end;


end.
